/*
 * Copyright 1996 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include </usr/include/sys/param.h>
#include <Assert.h>

#include "process.h"
#include "getRecord.h"
#include "filter.h"

static  int Filter_ValidRecord(char *addr);
static  int Filter_ValidBlockRecord(char *addr);
static void Filter_SetValid();
static void Filter_InitFilterFile(char *filterFile);


/* filter state */
static int filterFS = 0;
static int firstDev = -1;
static int lastDev = -1;
static int blockified;

static int valid[MAXCALLS+1];


/*
 *-----------------------------------------------------------------
 * Filter_Init --
 *
 * useRiver -- set to nonzero to use River (currently not available
 *             outside UCB)
 *
 * blocks   -- set to nonzero if input has been processed into blocks
 *
 * filterFile -- name of file containing filtering specification,
 *               set to NULL of "" to get all records.
 *               Filter file format is described below with the function
 *               Filter_InitFilterFile.
 *               
 *
 * numstreams -- number of input streams
 *
 * filelists  -- list of filenames for each stream
 *
 *-----------------------------------------------------------------
 */
void
Filter_Init(int useRiver, int blocks, char *filterFile,
	    int numstreams, char **filelists)
{
  GetRecord_Init(useRiver, numstreams, filelists);

  blockified = blocks;
  
  if (!blockified) {
    Filter_SetValid();
  }

  if ((filterFile == NULL) || (!strcmp(filterFile, ""))
      || (!(strcmp(filterFile, "none")))) {
    return;
  }

  Filter_InitFilterFile(filterFile);
}


/*
 *-----------------------------------------------------------------
 * Filter_NextRecord --
 *   copy record with lowest time into addr and return 1.
 *   if no more records, return 0.
 *   
 *-----------------------------------------------------------------
 */
int
Filter_NextRecord(char *addr)
{
  int ret;
  
  if (!blockified) {
    /* read until we get a valid record */
    while (((ret = GetRecord_NextRecord(addr)) != 0) &&
	   !(Filter_ValidRecord(addr)));
  }

  else {
    /* read until we get a valid record */
    while (((ret = GetRecord_NextRecord(addr)) != 0) &&
	   !(Filter_ValidBlockRecord(addr)));
  }

  return ret;
}



/*
 *-----------------------------------------------------------------
 * Filter_ValidRecord --
 *   check all applicable fields.
 *   if valid, return 1. otherwise return 0.
 *
 *-----------------------------------------------------------------
 */
static int
Filter_ValidRecord(char *addr)
{
  newHeader *hp = (newHeader *) addr;
  DevType *devp = (DevType *) (addr + sizeof(newHeader));

  
  /* always check for valid call */
  if ((valid[hp->ah_event] == -1 ) ||
      (valid[hp->ah_event] != hp->ah_len)) {
    fprintf(stderr,
	    "Filter_NextRecord: invalid call %d len %d\n",
	    hp->ah_event, hp->ah_len);
    exit(1);
  }


  if (filterFS) {
    if ((hp->ah_event != EXIT) &&
	(hp->ah_event != FORK) &&
	(hp->ah_event != VFORK) &&
	(hp->ah_event != SYNC) &&
	(hp->ah_event != REBOOT) &&
	(hp->ah_event != LSYNC)) {
      
      /* all other system calls have the format:
	 dev fileid otherArgs*/
      if ((*devp < firstDev) || (*devp > lastDev)) {
	return 0;
      }
    }
  }

  return 1;
}


/*
 *-----------------------------------------------------------------
 * Filter_ValidRecordBlock --
 *   check all applicable fields.
 *   if valid, return 1. otherwise return 0.
 *
 *-----------------------------------------------------------------
 */
static int
Filter_ValidBlockRecord(char *addr)
{
  newHeader *hp = (newHeader *) addr;
  DevType *devp = (DevType *) (addr + sizeof(newHeader));

  
  /* always check for valid call */
  ASSERT((hp->ah_event == OPEN) ||
	 (hp->ah_event == SREAD) ||
	 (hp->ah_event == SWRITE) ||
	 (hp->ah_event == TRUNC) ||
	 (hp->ah_event == UNLINK));

  ASSERT(hp->ah_len == sizeof(DevFidBlock));
  

  if (filterFS) {
    /* all system calls have the format:
       dev fileid otherArgs*/
    if ((*devp < firstDev) || (*devp > lastDev)) {
      return 0;
    }
  }

  return 1;
}


/*
 *-----------------------------------------------------------------
 * Filter_InitFilterFile --
 *
 * Current format of filter file:
 * Comment lines starting with # are ignored.
 *
 * First line must specify cluster.
 * Second line specifies which file system to keep.
 *
 * cluster res|ins
 * fs mail|root|tmp|home|bin
 *
 * For example:
 * cluster ins
 * fs mail
 *-----------------------------------------------------------------
 */
static void
Filter_InitFilterFile(char *filterFile)
{
  FILE *filep;
  char key[MAXPATHLEN];
  char value[MAXPATHLEN];
  int firstBin = -1;
  int lastBin = -1;
  int firstHome = -1;
  int lastHome = -1;
  int firstMail = -1;
  int lastMail = -1;
  int firstRoot = 0;
  int lastRoot = 100;
  int firstTmp = 2000;
  int lastTmp = 2999;
  char c;
  

  filep = fopen(filterFile, "r");

  if (filep == NULL) {
    perror ("Filter_InitFilterFile: open");
    fprintf(stderr, "Error opening file %s\n",
	    filterFile);
    exit(1);
  }

  while (fscanf(filep, "%s", key) == 1) {

    /* comment, skip to end of line */
    if (key[0] == '#') {
      while(((c = getc(filep)) != EOF)) {
	if (c == '\n') {
	  break;
	}
      }
      if (c == EOF) {
	break;
      }
      continue;
    }

    
    /* cluster */
    if (!(strcmp(key, "cluster"))) {

      if (fscanf(filep, "%s", value) != 1) {
	fprintf(stderr, "Error reading value for %s in %s\n",
		key, filterFile);
	exit(1);
      }
      if (!(strcmp(value, "res"))) {
	firstBin = 4052;
	lastBin = 4055;
	firstHome = 4029;
	lastHome = 4051;
	firstMail = 4056;
	lastMail = 4056;
      }
      else if (!(strcmp(value, "ins"))) {
	firstBin = 4209;
	lastBin = 4213;
	firstHome = 4101;
	lastHome = 4207;
	firstMail = 4219;
	lastMail = 4219;
      }
      else {
	fprintf(stderr, "Unknown value %s for %s\n",
		value, key);
      }
    } /* cluster */


    /* file system */
    if (!(strcmp(key, "fs"))) {

      if (fscanf(filep, "%s", value) != 1) {
	fprintf(stderr, "Error reading value for %s in %s\n",
		key, filterFile);
	exit(1);
      }
      filterFS = 1;
      

      if (!strcmp(value, "home")) {
	if ((firstHome == -1) || (lastHome == -1)) {
	  fprintf(stderr, "cluster must be set before file system in %s\n",
		  filterFile);
	  exit(1);
	}
	firstDev = firstHome;
	lastDev = lastHome;
      }
      else if (!strcmp(value, "bin")) {
	if ((firstHome == -1) || (lastHome == -1)) {
	  fprintf(stderr, "cluster must be set before file system in %s\n",
		  filterFile);
	  exit(1);
	}
	firstDev = firstBin;
	lastDev = lastBin;
      }
      else if (!strcmp(value, "root")) {
	if ((firstHome == -1) || (lastHome == -1)) {
	  fprintf(stderr, "cluster must be set before file system in %s\n",
		  filterFile);
	  exit(1);
	}
	firstDev = firstRoot;
	lastDev = lastRoot;
      }
      else if (!strcmp(value, "mail")) {
	if ((firstHome == -1) || (lastHome == -1)) {
	  fprintf(stderr, "cluster must be set before file system in %s\n",
		  filterFile);
	  exit(1);
	}
	firstDev = firstMail;
	lastDev = lastMail;
      }
      else if (!strcmp(value, "tmp")) {
	if ((firstHome == -1) || (lastHome == -1)) {
	  fprintf(stderr, "cluster must be set before file system in %s\n",
		  filterFile);
	  exit(1);
	}
	firstDev = firstTmp;
	lastDev = lastTmp;
      }
      else {
	fprintf(stderr, "unknown file system type %s\n", value);
	exit(1);
      }

    } /* file system */
  }
}

  

/*-----------------------------------------------------------------*/

static void
Filter_SetValid()
{
  int i;

  for (i=0; i <= MAXCALLS; i++) {
    valid[i] = -1;
  }

  valid[1]   = 0;            /* EXIT   */
  valid[2]   = FORKARGSIZE;  /* FORK   */
  valid[3]   = IOARGSIZE;    /* SREAD  */
  valid[4]   = IOARGSIZE;    /* SWRITE */
  valid[5]   = OPENARGSIZE;  /* OPEN   */
  valid[6]   = CLOSEARGSIZE; /* CLOSE  */
  valid[8]   = OPENARGSIZE;  /* CREAT  */
  valid[9]   = DEVFIDSIZE;   /* LINK   */
  valid[10]  = DEVFIDSIZE;   /* UNLINK */
  
  valid[11]  = DEVFIDSIZE;   /* EXECV  */
  valid[12]  = DEVFIDSIZE;   /* CHDIR  */
  valid[15]  = DEVFIDSIZE;   /* CHMOD  */
  valid[16]  = DEVFIDSIZE;   /* CHOWN  */
  valid[19]  = -1;           /* LSEEK  */
  
  valid[21]  = DEVFIDSIZE;   /* SMOUNT */
  valid[22]  = DEVFIDSIZE;   /* UMOUNT */
  valid[30]  = DEVFIDSIZE;   /* UTIME  */

  valid[33]  = DEVFIDSIZE;   /* ACCESS */
  valid[36]  = 0;            /* SYNC   */
  valid[38]  = DEVFIDSIZE;   /* STAT   */
  valid[40]  = DEVFIDSIZE;   /* LSTAT  */

  valid[55]  = 0;            /* REBOOT  */
  valid[56]  = DEVFIDSIZE;   /* SYMLINK */
  valid[58]  = DEVFIDSIZE;   /* RDLINK  */
  valid[59]  = DEVFIDSIZE;   /* EXECVE  */

  valid[61]  = DEVFIDSIZE;   /* CHROOT  */
  valid[66]  = FORKARGSIZE;  /* VFORK   */
  
  valid[71]  = DEVFIDSIZE;   /* MMAP    */
  valid[73]  = DEVFIDSIZE;   /* MUNMAP  */

  valid[90]  = -1;           /* DUP2    */
  valid[92]  = DEVFIDSIZE;   /* FSTAT   */
  valid[95]  = DEVFIDSIZE;   /* FSYNC   */
  valid[120] = IOARGSIZE;    /* READV   */
  valid[121] = IOARGSIZE;    /* WRITEV  */
  valid[123] = DEVFIDSIZE;   /* FCHOWN  */
  valid[124] = DEVFIDSIZE;   /* FCHMOD  */
  valid[128] = DEVFIDSIZE;   /* RENAME  */
  valid[129] = TRUNCARGSIZE; /* TRUNC   */
  valid[130] = TRUNCARGSIZE; /* FTRUNC  */
  valid[136] = DEVFIDSIZE;   /* MKDIR   */
  valid[137] = DEVFIDSIZE;   /* RMDIR   */
  valid[155] = LOCKARGSIZE;  /* LOCKF   */
  valid[178] = 0;            /* LSYNC   */
  valid[195] = DEVFIDSIZE;   /* GETDIRENTRIES */
  valid[198] = DEVFIDSIZE;   /* VFSMOUNT */
  valid[235] = DEVFIDSIZE;   /* GETACL   */
  valid[236] = DEVFIDSIZE;   /* FGETACL  */
  valid[237] = DEVFIDSIZE;   /* SETACL   */
  valid[238] = DEVFIDSIZE;   /* FSETACL  */
  valid[249] = DEVFIDSIZE;   /* GETACCESS */
  valid[250] = -1;           /* FSCTL    */
  valid[267] = DEVFIDSIZE;   /* TSYNC    */
  valid[272] = DEVFIDSIZE;   /* FCHDIR   */
}


