/*
 * Copyright 1996 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include </usr/include/sys/param.h>
#include <unistd.h>

#include <Assert.h>
#include <process.h>
#include <getRecord.h>


/* CONFIGURE THESE WITH CORRECT PATH NAMES FOR YOUR ENVIRONMENT */

#define UNCOMPRESS_COMMAND "gzip -dc"
#define CAT_COMMAND "cat"

/* River currently not available outside UCB. */
#define RIVER_UNCOMPRESS_COMMAND "/home/cs/drew/src/rivers/rivers/drew/traceco"
#define RIVER_CAT_COMMAND "/home/cs/drew/src/rivers/rivers/drew/traceco-nocompress"

/* END CONFIGURATION */


/*
 *-----------------------------------------------------------------
 * GetRecord --
 * 
 * This module is intended to provide buffered and sorted access
 * to several streams of input.
 *
 * Each input stream must be sorted internally.
 * Each stream is represented by a file containing the names
 * of all trace files for that stream in order.
 *
 * GetRecord will choose the record that has the lowest time value
 * among all records at the heads of streams.
 *
 * All records must begin with a newHeader.
 * The body of the record is ignored. The length the record
 * is determined solely by header.ah_len, so this works for both
 * regular traces and blockified traces.
 *-----------------------------------------------------------------
 */

#define RBUFFSIZE 4096


typedef struct {
  FILE *currentStreamp;
  int   active;
  int   currentTime;
  newHeader header;
  char *readbuf;
  char *readbp;
  char *rendp;
} StreamStruct;

static void GetRecord_SwitchFiles(int stream);
static void GetRecord_InitHeader(int stream);
static void GetRecord_ReadBuffer(int stream);
static void GetRecord_ReadRecord(int stream, char *recordp);
static  int GetRecord_NextRecordFile(char *addr);

static void GetRecord_InitStdinBuffer (StreamStruct *state);
static  int GetRecord_NextRecordStdin (char *addr);
static  int GetRecord_ReadStdio(StreamStruct *state, char *addr, int size);


/* getRecord state */
static  int numStreams;
static  int numActiveStreams;
static char **filelist_filenames;
static FILE **filelistp;
static  int useRivers;
static  int useStdin;

static StreamStruct *streams;



/*
 *-----------------------------------------------------------------
 * GetRecord_Init --
 *
 * rivers -- set to nonzero to use River (currently not available
 *           outside UCB).
 *
 * numstreams -- number of input streams
 *
 * filelists -- a list of filenames representing streams.
 *              Each stream contains a list of trace file names.
 *              Each trace file contains an ordered set of records.
 *
 *-----------------------------------------------------------------
 */
void
GetRecord_Init(int rivers, int numstreams, char **filelists)
{
  int i;

  if (numstreams == 0) {
    useStdin = 1;
    streams = (StreamStruct *) calloc (1, sizeof(StreamStruct));
    GetRecord_InitStdinBuffer (streams);
  }

  else {
    useStdin = 0;
    useRivers = rivers;
    numStreams = numstreams;
    numActiveStreams = numstreams;
    filelist_filenames = filelists;

    filelistp = (FILE **) calloc (numStreams, sizeof(FILE *));
    streams = (StreamStruct *) calloc (numStreams, sizeof(StreamStruct));
  
    ASSERT(filelistp && streams);

  
    /* open all file name files */
    for (i=0; i < numstreams; i++) {
      filelistp[i] = fopen(filelist_filenames[i], "r");
      if (filelistp[i] == NULL) {
	perror("GetRecord_Init: open");
	fprintf(stderr, "error opening stream file %s\n",
		filelist_filenames[i]);
	exit(1);
      }
      streams[i].readbuf = malloc (RBUFFSIZE);
      ASSERT(streams[i].readbuf);
      GetRecord_ReadBuffer(i);
      streams[i].currentStreamp = NULL;
      streams[i].active = 1;
      GetRecord_SwitchFiles(i);
      GetRecord_InitHeader(i);
    }
  }
}


/*
 *-----------------------------------------------------------------
 * GetRecord_SwitchFiles --
 *   open the next file for this input stream.
 *
 *   if no more files, set active to 0.
 *
 *-----------------------------------------------------------------
 */
static void
GetRecord_SwitchFiles(int which)
{
  char filename[MAXPATHLEN];
  char cmdLine[MAXPATHLEN];

  
  if (streams[which].currentStreamp != NULL) {
    if (pclose(streams[which].currentStreamp)) {
      perror("pclose");
      exit(1);
    }
  }

  /* read next input file name */
  if (fscanf(filelistp[which], "%s", filename) != 1) {
    if (ferror(filelistp[which])) {
      perror("fscanf");
      fprintf(stderr, "error reading %s\n", filelist_filenames[which]);
      exit(1);
    }
    /* no more input files for this stream */
    if (feof) {
      streams[which].active = 0;
      numActiveStreams--;
      return;
    }
  }

  /* open next input file */
  if (!strcmp(".gz", &(filename[strlen(filename) - 3]))) {

    if (useRivers) {
     sprintf(cmdLine, "%s %s", RIVER_UNCOMPRESS_COMMAND, filename);

    }
    else {
      sprintf(cmdLine, "%s %s", UNCOMPRESS_COMMAND, filename);
    }
  }
  else {

    if (useRivers) {
      sprintf(cmdLine, "%s %s", RIVER_CAT_COMMAND, filename);
    }
    else {
      sprintf(cmdLine, "%s %s", CAT_COMMAND, filename);
    }
  }
  
  streams[which].currentStreamp = popen(cmdLine, "r");

  if (streams[which].currentStreamp == NULL) {
    perror("popen");
    fprintf(stderr, "error opening %s\n", filename);
    exit(1);
  }
  fprintf(stderr, "reading from %s\n", filename);
}



/*
 *-----------------------------------------------------------------
 * GetRecord_InitHeader --
 *   init header and currentTime for stream
 *
 *
 *-----------------------------------------------------------------
 */
static void
GetRecord_InitHeader(int which)
{
  char *addr;
  int size;
  
  /* read first header */
  addr = (char *) &(streams[which].header);
  size = sizeof(newHeader);
  while ((streams[which].rendp - streams[which].readbp) < size) {
    memcpy(addr, streams[which].readbp,
	   streams[which].rendp - streams[which].readbp);
    size -= (streams[which].rendp - streams[which].readbp);
    addr += (streams[which].rendp - streams[which].readbp);

    GetRecord_ReadBuffer(which);
    if (!streams[which].active) {
      if (size != sizeof(newHeader)) {
	fprintf(stderr, "Incomplete header. init header. size %d\n", size);
	exit(1);
      }
      else {
	return;
      }
    }
  }
  memcpy(addr, streams[which].readbp, size);
  streams[which].readbp+=size;

  /* update current time for this stream */
  streams[which].currentTime = streams[which].header.ah_time;
  if (streams[which].currentTime < 0) {
    streams[which].currentTime = 0 - streams[which].currentTime;
  }
}



/*
 *-----------------------------------------------------------------
 * GetRecord_ReadBuffer --
 *    read in next bufferfull.
 *    switch files if eof.
 *
 *    return when >= 1 bytes have been read
 *    or no more input.
 *
 * switchFiles sets active to 0 if no more input.
 *-----------------------------------------------------------------
 */
static void
GetRecord_ReadBuffer(int which)
{
  int numbytes = 0;


  while (streams[which].active && (numbytes <= 0)) {
    numbytes = fread(streams[which].readbuf, 1, RBUFFSIZE,
		     streams[which].currentStreamp);
    
    if (numbytes <= 0) {
      if (ferror(streams[which].currentStreamp)) {
	perror("fread");
	exit(1);
      }
      if (feof(streams[which].currentStreamp)) {
	GetRecord_SwitchFiles(which);
      }
    }
  }

  streams[which].readbp = streams[which].readbuf;
  streams[which].rendp = streams[which].readbuf + numbytes;
}


/* faster than memcpy for small copies */
static void
MEMCPY(char *dest, char *src, int totalbytes)
{
  int i, numbytes;
  int *srcp = (int *) src;
  int *destp = (int *) dest;

  ASSERT(!(totalbytes % sizeof(int)));
  for (i=0, numbytes=0; numbytes < totalbytes; i++, numbytes += sizeof(int)) {
    destp[i] = srcp[i];
  }
}



/*
 *-----------------------------------------------------------------
 * GetRecord_ReadRecord --
 *   copy record into addr, update nextTime.
 *   call ReadBuffer if buffer empty.
 *
 *-----------------------------------------------------------------
 */
static void
GetRecord_ReadRecord(int which, char *recordp)
{
  int size;
  char *addr;
  newHeader *hp = (newHeader *) recordp;

  /* copy header to addr */
  *hp = streams[which].header;
  
  /* read rest of record */

  addr = recordp + sizeof(newHeader);
  size = hp->ah_len;
  ASSERT(size <= MAX_TRACE_RECORD_SIZE);
  while ((streams[which].rendp - streams[which].readbp) < size) {
    memcpy(addr, streams[which].readbp, (streams[which].rendp - streams[which].readbp));
    size -= (streams[which].rendp - streams[which].readbp);
    addr += (streams[which].rendp - streams[which].readbp);

    GetRecord_ReadBuffer(which);
    if (!streams[which].active) {
      fprintf(stderr, "Incomplete record\n");
      exit(1);
    }
  }
  MEMCPY(addr, streams[which].readbp, size);
  streams[which].readbp+=size;


  /* read next header */

  addr = (char *) &(streams[which].header);
  size = sizeof(newHeader);
  while ((streams[which].rendp - streams[which].readbp) < size) {
    memcpy(addr, streams[which].readbp, (streams[which].rendp - streams[which].readbp));
    size -= (streams[which].rendp - streams[which].readbp);
    addr += (streams[which].rendp - streams[which].readbp);

    GetRecord_ReadBuffer(which);
    if (!streams[which].active) {
      if (size != sizeof(newHeader)) {
	fprintf(stderr, "Incomplete header. size %d\n", size);
	exit(1);
      }
      else {
	return;
      }
    }
  }
  MEMCPY(addr, streams[which].readbp, size);
  streams[which].readbp+=size;

  /* update current time for this stream */
  
  streams[which].currentTime = streams[which].header.ah_time;
  if (streams[which].currentTime < 0) {
    streams[which].currentTime = 0 - streams[which].currentTime;
  }
}



/*
 *-----------------------------------------------------------------
 * GetRecord_NextRecordFile --
 *   copy record with lowest time into addr and return 1.
 *   if no more records, return 0.
 *
 *   in the event of a tie, lower streams have priority.
 *-----------------------------------------------------------------
 */
static int
GetRecord_NextRecordFile(char *addr)
{
  int i;
  int lowestTime = -1;
  int lowestStream = -1;

  /* find stream with lowest time */
  for (i=0; i < numStreams; i++) {
    if (streams[i].active) {
    
      if (lowestTime == -1) {
	lowestTime = streams[i].currentTime;
	lowestStream = i;
      }
    
      else if (streams[i].currentTime < lowestTime) {
	lowestTime = streams[i].currentTime;
	lowestStream = i;
      }
    }
  }    

  if (numActiveStreams <= 0) {
    return GETRECORD_EOF;
  }
  
  ASSERT((lowestStream >= 0) && (lowestStream < numStreams));
  
  /* read record into addr */
  GetRecord_ReadRecord(lowestStream, addr);

  return GETRECORD_NOT_EOF;
}


/*
 *-----------------------------------------------------------------
 * GetRecord_NextRecord --
 *   copy record with lowest time into addr and return 1.
 *   if no more records, return 0.
 *
 *   in the event of a tie, lower streams have priority.
 *-----------------------------------------------------------------
 */
int
GetRecord_NextRecord(char *addr)
{
  if (useStdin) {
    return GetRecord_NextRecordStdin(addr);
  }
  else {
    return GetRecord_NextRecordFile(addr);
  }
}
    


/*
 *-----------------------------------------------------------------
 * GetRecord_InitStdinBuffer --
 *
 *-----------------------------------------------------------------
 */
static void
GetRecord_InitStdinBuffer (StreamStruct *state)
{
  state->readbuf = (caddr_t) malloc(RBUFFSIZE);
  
  if (state->readbuf == NULL) {
    fprintf(stderr, "GetRecord_InitStdinBuffer. Out of memory\n");
    exit(1);
  }

  state->rendp = state->readbuf + RBUFFSIZE;

  state->readbp = state->rendp;
}


static int
GetRecord_NextRecordStdin (char *addr)
{
  newHeader *hp;
  int ret;
  
  ret = GetRecord_ReadStdio(streams, addr, sizeof(newHeader));
  if (ret == GETRECORD_EOF) {
    return GETRECORD_EOF;
  }
  
  hp = (newHeader *) addr;
  
  ret = GetRecord_ReadStdio(streams, addr + sizeof(newHeader), hp->ah_len);
  ASSERT( ret != GETRECORD_EOF);

  return GETRECORD_NOT_EOF;
}


/*
 *-----------------------------------------------------------------
 * GetRecord_ReadStdio --
 *
 * return 1 when out of input
 *
 *-----------------------------------------------------------------
 */
static int
GetRecord_ReadStdio(StreamStruct *state, char *addr, int size)
{
  int bytes;
  int readsize;
  int bytesRead;
  
  ASSERT(size < RBUFFSIZE);


  /* read in new bufferful */
  if ((state->rendp - state->readbp) < size) {

    /* copy rest of buffer into addr */
    memcpy(addr, state->readbp, (state->rendp - state->readbp));
    addr += (state->rendp - state->readbp);
    size -= (state->rendp - state->readbp);

    /* read in new bufferful */
    state->readbp = state->readbuf;
    readsize = RBUFFSIZE;
    bytesRead = 0;

    while (bytesRead < size) {
      bytes = read(0, state->readbp, readsize);
      if (bytes < 0) {
	perror("GetRecord_ReadStdio error");
	exit(1);
      }
      if (bytes == 0) {
	return GETRECORD_EOF;
      }
      readsize -= bytes;
      bytesRead += bytes;
      state->readbp += bytes;
    }
    state->readbp = state->readbuf;
    state->rendp = state->readbp + bytesRead;
  }

  /* copy bytes into callers addr, update current buffer pointer */
  memcpy(addr, state->readbp, size);
  state->readbp+=size;

  return GETRECORD_NOT_EOF;
}
