/*
 * Copyright 1996 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */
/*
 *-----------------------------------------------------------------
 * 
 * TRACE FILE FORMAT
 *   each day of traces is stored in a separate file,
 *   the file only contains processed trace records, stored in
 *   big endian binary format
 *
 * TRACE RECORD FORMAT
 *   Every trace record begins with a header. Depending on the
 *   call type, the header may be followed by arguments.
 *
 *-----------------------------------------------------------------
 */

#ifndef _PROCESS_H
#define _PROCESS_H

#include <sys/types.h>

#define MAXCALLS 273
#define UNKNOWN -9
#define MAX_TRACE_RECORD_SIZE 100

#define MAX_DEV 4345


/* record header, 20 bytes */
typedef struct  {
  long ah_time;	          /* date/time (tv_sec of timeval)
			     a negative time indicates unknown time
			     between clock resets */
  long ah_host;           /* host ID */
  long ah_pid;		  /* process ID */
  long ah_uid;		  /* user ID */
  u_short ah_event;	  /* syscall number */
  u_short ah_len;	  /* length of following arguments */
} newHeader;



/*
 *-----------------------------------------------------------------
 *
 * ARGUMENT STRUCTURES
 *
 *-----------------------------------------------------------------
 */
#define RAW_PATH_ARGS_SIZE ((3 * sizeof(int)) + sizeof(short))
#define PATH_ARGS_SIZE (sizeof(FIDType) + sizeof(DevType) + sizeof(short))
#define DEVFIDSIZE (sizeof(DevType) + sizeof(FIDType))
#define FORKARGSIZE (2 * sizeof(int))
#define CLOSEARGSIZE (sizeof(DevType) + sizeof(FIDType) + sizeof(int))
#define TRUNCARGSIZE (sizeof(DevType) + sizeof(FIDType) + sizeof(int))
#define OPENARGSIZE  (sizeof(DevType) + sizeof(FIDType) + 10 * sizeof(int))
#define IOARGSIZE (sizeof(DevType) + sizeof(FIDType) + 2 * sizeof(int))
#define LOCKARGSIZE (sizeof(DevType) + sizeof(FIDType) + 2 * sizeof(int))

typedef u_long DevType;   /* file system (device) identifier */
typedef long FIDType;     /* unique file id */


typedef struct {
  DevType dev;            /* file system (device) number */
  FIDType fid;            /* unique file id */
} DevFid;




/* close modes */
#define CLOSE_HOST   1    /* final close for file on this host */
#define CLOSE_PS     2    /* final close for file for this process */
#define CLOSE_DESC   3    /* any close on this descriptor */

/* file system types */
#define FSTYPE_UFS 1
#define FSTYPE_NFS 2
#define FSTYPE_DUX 3

/* lockf functions */
#define LOCKF_ULOCK   0      /* unlock a region */
#define LOCKF_LOCK    1      /* lock a region */
#define LOCKF_TLOCK   2      /* test and lock a region */
#define LOCKF_TEST    3      /* test region for lock */

/* file types */
#define FTYPE_REG 1          /* regular file */
#define FTYPE_DIR 2          /* directory */
#define FTYPE_SYMLINK 5      /* symbolic link */
#define FTYPE_EMPTYDIR 10    /* empty directory */

/* file open modes.
   note we can't use those defined in fcntl.h because they're different
   on different operating systems */
#define RDONLYMODE 0000000
#define WRONLYMODE 0000001
#define ACCMODE   0000003
#define CREATMODE 0000400
#define TRUNCMODE 0001000
#define APPENDMODE 0000010

#define SYNC_MODE    0100000    /* Synchronous writes */
#define NDELAY_MODE  0000004    /* Non-blocking open and/or I/O */
#define SYNCIO_MODE  SYNC_MODE    /* Do write through caching */
#define FSYNCIO_MODE SYNC_MODE    /* Do write through caching */


/* Values for cmd used by fcntl() */
#define CMD_DUPFD       0       /* Duplicate fildes */
#define CMD_GETFD       1       /* Get file descriptor flags */
#define CMD_SETFD       2       /* Set file descriptor flags */
#define CMD_GETFL       3       /* Get file status flags */
#define CMD_SETFL       4       /* Set file status flags */
#define CMD_GETLK       5       /* Get file lock */
#define CMD_SETLK       6       /* Set file lock */
#define CMD_SETLKW      7       /* Set file lock and wait */



/* calls that have no arguments:
 * EXIT, SYNC, REBOOT, LSYNC
 */


/* argument structure for:
 * MUNMAP, FSTAT, FSYNC, FCHDIR, FCHOWN, FCHMOD, GETDIRENTRIES,
 * FGETACL, FSETACL, MMAP, TSYNC, SMOUNT, UNLINK, EXECV, CHDIR,
 * CHMOD, CHOWN, UMOUNT, UTIME, ACCESS, STAT, LSTAT, RDLINK,
 * EXECVE, CHROOT, MKDIR, RMDIR, GETACL, SETACL, GETACCESS,
 * LINK, RENAME, SYMLINK, VFSMOUNT
 * is DevFid.
 */

/* argument structure for:
 * TRUNC, FTRUNC
 */
typedef struct {
  DevFid devfid;
  int size;         /* new file size */
} TruncArgs;



/* argument structure for:
 * CLOSE
 */
typedef struct {
  DevFid devfid;
  int closemode;   /* one of the close modes listed above */
} CloseArgs;


/* argument structure for:
 * SREAD, SWRITE, READV, WRITEV
 */
typedef struct {
  DevFid devfid;
  int offset;
  int bytes;
} ReadWriteArgs;


/* argument structure for:
 * LOCKF
 */
typedef struct {
  DevFid devfid;
  int lockf_function; /* one of the lockf functions listed above */  
  int size;           /* size of region from current offset to apply lock to */
} LockfArgs;



/* argument structure for:
 * FORK, VFORK
 * childFlag is 1 if this process is the child and 0 otherwise
 * pid is parent process id if this is the child and child process id
 * if this is the parent
 */
typedef struct {
  int pid;
  int childFlag;
} ForkArgs;

/* argument structure for:
 * OPEN, CREAT
 *
 * NOTA BENE: files that are opened in create mode may or may not be new
 * files. The kernel does not accurately record all information for these
 * files, therefore, these files have uid, size, and nlinks set to UNKNOWN.
 */
typedef struct {
  DevFid devfid;
  int fd;          /* file descriptor */
  int ftype;       /* file type, one of the FTYPEs defined above */
  int fstype;      /* file system type, one of the FSTYPEs defined above */
  int uid;         /* user id of the file's owner */
  int size;        /* file size in bytes */
  int nlink;       /* number of hard links */
  int ctime;       /* last change time */
  int mtime;       /* last modification time */
  int atime;       /* last access time */
  int mode;        /* open mode, one of the modes listed above */
} OpenArgs;

/* calls that should not be seen in processed traces:
 * LSEEK, DUP, DUP2, FCNTL, FSCTL
 */


#define EXIT 1
#define FORK 2
#define SREAD 3
#define SWRITE 4
#define OPEN 5
#define CLOSE 6
#define CREAT 8
#define LINK 9
#define UNLINK 10
#define EXECV 11
#define CHDIR 12
#define CHMOD 15
#define CHOWN 16
#define LSEEK 19
#define SMOUNT 21
#define UMOUNT 22
#define UTIME 30
#define ACCESS 33
#define SYNC 36
#define STAT 38
#define LSTAT 40
#define DUP 41
#define REBOOT 55
#define SYMLINK 56
#define RDLINK 58
#define EXECVE 59
#define CHROOT 61
#define FCNTL 62
#define VFORK 66
#define MMAP 71
#define MUNMAP 73
#define DUP2 90
#define FSTAT 92
#define FSYNC 95
#define READV 120
#define WRITEV 121
#define FCHOWN 123
#define FCHMOD 124
#define RENAME 128
#define TRUNC 129
#define FTRUNC 130
#define MKDIR 136
#define RMDIR 137
#define LOCKF 155
#define LSYNC 178
#define GETDIRENTRIES 195
#define VFSMOUNT 198
#define GETACL 235
#define FGETACL 236
#define SETACL 237
#define FSETACL 238
#define GETACCESS 249
#define FSCTL 250
#define TSYNC 267
#define FCHDIR 272

#endif
